package no.rwr.x;
import gnu.gcj.RawData;



/** The format, depth and offset attributes of an XImage determines
    how a bitfield is encoded in a raster image. However, it does not
    determine how a color is encoded into a bitfield. */

/*

  bytes_per_line, lines must be byte aligned

  bytes_per_plane = height * bytes_per_line

  planeOffset = planeN * bytes_per_plane; // only XYPIXMAP
  // or
  planeOffset = 0;                        // ZPIXMAP only has one chunky plane

  bits_per_pixel = 1  // for XYPIXMAP

  data[planeOffset + (y * bytes_per_line) + round_to_unit(x * bits_per_pixel)]


  Manipulating XYPIXMAP:
    Find addresses/offsets of bitmap_units. <--- (plane, y,) x:
      // Downshift and clear lower bits
      // eg: bitmap_unit = 8:
      //   x >> 3
      // eg: bitmap_unit = 16:
      //   x >> 3 & (~0x01)    ...or... (x>>4)<<1 
      // eg: bitmap_unit = 32: ...or... (x>>5)<<2
      //   x >> 3 & (~0x03) 

      // generalization:
      //   unitIndexBits = log2(bitmap_unit)   // eg. 32 ---> 5
      //   bytesPerUnitShift = unitIndexBits-3
      //   unitOffset     = (x>>unitIndexBits)   // useful? array access?
      //   unitByteOffset = (x>>unitIndexBits)<<bytesPerUnitShift
      //   unitIndexMask = bitmap_unit-1;   // eg. 0x20 ---> 0x1f
      //   bitIndex      = x & unitIndexMask
      
      // eg: If bitmap_unit is 16, i.e. bits are grouped in units of 16,
      // then the 4 (log2(16)) lower bits of x is the bitIndex within the
      // unit. The other bits (x>>4) is the offset to the unit. One unit
      // consists of 2 bytes (2^(4-3)).
      //
      // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx           x
      //                             xxxx           bitIndex
      // xxxxxxxxxxxxxxxxxxxxxxxxxxxx               unitOffset
      // xxxxxxxxxxxxxxxxxxxxxxxxxxxx0              unitByteOffset
      //                             ----           unitIndexBits
      //                              ---           3: x to byte shift (x>>3)
      //                             -              bytesPerUnitShift
      //
      // perhaps:
      //   bitIndex = x & unitIndexMask;
      //   unitByteOffset = (x ^ bitIndex)>>3;

      // or lookup table both for unitOffset AND bitmask
      //   addr = unitByteOffset[x];
      //   mask = bitMask[x]

      // or mix:
      //   addr = roundUnit[x<<3];
      //   mask = bitMask[x & 0x1f];   // ((1<<5)-1) == 0x1f

      // Whichever is faster...

    Find bitIndex  <--- x
    Grab bitmap_unit.
    Process unit:


      Normalize unit
      Manipulate:
        Create bitMask from bitIndex (once)
	Use bitmask on normalized unit
      Denormalize unit

      // OR (optimization)
      
      Lookup bitmask from precreated table using bitIndex (once)
      Use bitmask on non-normalized unit
      

    Put bitmap_unit.


  Writing a whole unit can be optimized to a set of bit shifts + AND + OR.


  bits_per_pixel: number of bits per pixel (length of pixel bitfield)
  for Z-format pixmap.

  ?? how does depth relate to bits_per_pixel?

  bits_per_pixel does not need to be divisable by 8. (for Z-Pixmap),
  i.e. all pixels in a Z-Pixmap do not start at a unique byte address.

  bits_per_pixel is either 1, 4, 8, 16, 24 or 32 (traditional assumption)
  
  bitmap_unit is divisable by 8. Traditionally: 8, 16 or 32 bits

  
  ex: 

  bytes_per_line = 4
  depth = 2
        
         12
        |<>| bits_per_pixel = 2 
        [xx]

0x0000  [00][00][00][xx] [xx][xx][xx][xx] [xx][xx][xx][xx] [xx][xx][xx][xx]
0x0004  [00][00][00][xx] [xx][xx][xx][xx] [xx][xx][xx][xx] [xx][xx][xx][xx]



  ex:
 
  byte_per_line = 6
  depth = 24

                   1111111 11122222
         12345678 90123456 78901234 
        |<------------------------>| bits_per_pixel = 24
        [xxxxxxxx xxxxxxxx xxxxxxxx]

0x0000  [xxxxxxxx xxxxxxxx xxxxxxxx] [xxxxxxxx xxxxxxxx xxxxxxxx]
0x0006  [xxxxxxxx xxxxxxxx xxxxxxxx] [xxxxxxxx xxxxxxxx xxxxxxxx]


normalized byte order:


     0-7      8-15     16-23    24-31
                111111 22221111 33222222
     76543210 54321098 32109876 10987654
    [xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx]
     76543210 76543210 76543210 76543210
     

...abcdefghijklmnop...

BitmapBitOrder MSBFirst:
    [abcdefghijklmnop]   16bits
     ^MSBit         ^LSBit
     ^First (leftmost on display)
     <MSByte><LSByte>

    ByteOrder MSBFirst:
        [abcdefgh] [ijklmnop]   8bits
        ^MSByte    ^LSByte
	^First (in memory)

    ByteOrder LSBFirst:
        [ijklmnop] [abcdefgh]   8bits
        ^LSByte    ^MSByte
	^First (in memory)

BitmapBitOrder LSBFirst:
    [ponmlkjihgfedcba]   16bits
     ^MSBit         ^LSBit
                    ^First (leftmost on display)
     <MSByte><LSByte>

    ByteOrder MSBFirst:
        [ponmlkji] [hgfedcba]   8bits
        ^MSByte    ^LSByte
	^First (in memory)

    ByteOrder LSBFirst:
        [hgfedcba] [ponmlkji]   8bits
        ^LSByte    ^MSByte
	^First (in memory)

ByteOrder: MSBFirst; look at the last byte in unit for the first bit
ByteOrder: LSBFirst; look at the first byte in the unit for the first bit





byte-order is the same on the whole display.

 */


/* 
   Concerning pixmaps: A pixmap is bound to the screen on which it was
   created.



   How does this relate to ScreenFormat?


   What is the difference between bitmap_unit and bitmap_pad?
   (The client can define the pad, but unit is server defined?)
 */

/* Depth can be less than bits_per_pixel. For example:
   depth = 24                 // Servers only as an indicator that
                              // 2^24 unique colors can be displayed.
   bits_per_pixel = 32

*/

public class XImage {

    /** This object reference points to the data, hindering garbage
	collection of the data. */
    Object dataRef;


    //    /** @deprecated no need to store visual. The XImage structure
    //	contains all the necessary information.  */
    //    Visual visual;
    
    public static final int XYBITMAP_FORMAT = 0;
    public static final int XYPIXMAP_FORMAT = 1;
    public static final int ZPIXMAP_FORMAT  = 2;

    public static final int LEAST_SIGNIFICANT_B_FIRST_ORDER  = 0;
    public static final int MOST_SIGNIFICANT_B_FIRST_ORDER   = 1;

    public XImage(Visual visual, int depth, int format, int xoffset,
		  int width, int height, int bitmapPad,
		  int bytesPerLine) {
	// FIXME: set visual?
	init(visual, depth, format, xoffset, width, height,
	     bitmapPad, bytesPerLine);
    }

    public native void init(Visual visual, int depth, int format, int xoffset,
		     int width, int height, int bitmapPad,
		     int bytesPerLine);
    
    public XImage(Visual visual, int width, int height) {
	this(visual, width, height,
	     true // Automatically allocate memory
	     );
    }

    /** 
	@param allocate specifies whether to automatically allocate
	       memory for the image. It is possible to create the data
	       array elsewhere, so that we can for instance use a
	       DataBufferUShort as data. Ie. not limit ourself to byte
	       arrays. This is done by passing false and calling a
	       setData method manually after creation. */
    public XImage(Visual visual, int width, int height, boolean allocate) {
	init(visual, width, height);

	if (allocate) {
	    /* Now that Xlib has figured out the appropriate bytes per line,
	       we can allocate memory for the image.  */
	    // FIXME: What about formats with several layers/bands?
	    byte[] data = new byte[getBytesPerLine()*height];

	    /* FIXME: A bit more thought should be put into the
	       allocation and setting of the data. We should be able
	       to create a DataBuffer (and WritableRaster) describing
	       the data. This is can clearly not be done here, since
	       the no.rwr.x package does not recognize the existance
	       of java.awt. However, the no.rwr.xutil package should
	       be able to do it based on the facilities provided by
	       this class. */
	    setData(data, 0);
	}
    }

    private native void init(Visual visual, int width, int height);

    /**
       @param offset the index of the first actual data element in the array.
    */
    public void setData(byte[] data, int offset) {
	dataRef = data;
	internalSetData(data, offset);
    }

    /**
       @param offset the index of the first actual data element in the array.
       Note: this is not a byte offset.
    */
    public void setData(short[] data, int offset) {
	dataRef = data;
	internalSetData(data, offset);
    }

    /**
       @param offset the index of the first actual data element in the array.
       Note: this is not a byte offset.
    */
    public void setData(int[] data, int offset) {
	dataRef = data;
	internalSetData(data, offset);
    }

    private native void internalSetData(byte[] data, int offset);
    private native void internalSetData(short[] data, int offset);
    private native void internalSetData(int[] data, int offset);
    
    protected native void finalize();

    boolean ownsData = false;
    RawData structure = 0;

    
    //    /** @deprecated no need to store visual. The XImage structure
    //	contains all the necessary information.  */
    //    public Visual getVisual() {
    //	return visual;
    //    }

    public final native int getWidth();
    public final native int getHeight();
    public final native int getDepth();
    public final native int getFormat();
    public final boolean isZPixmapFormat() {
	return getFormat() == ZPIXMAP_FORMAT;
    } 


    /** The xoffset is used to avoid the need of shifting the
        scanlines into place. */
    public final native int getXOffset();


    public native final int getBytesPerLine();
    public native final int getBitsPerPixel();

    public native final int getImageByteOrder();
    public native final int getBitmapBitOrder();
    public native final int getBitmapUnit();
    public native final int getBitmapPad();


    // True/Direct Color specific:
    public native int getRedMask();
    public native int getGreenMask();
    public native int getBlueMask();


    // These are slow, don't use them, execpt as a fall-back:
    public native final void setPixel(int x, int y, int pixel);


    /* @deprecated */
    public final Object getData() {
	return dataRef;
    }

    public String toString() {
	String format;
	switch(getFormat()) {
	case ZPIXMAP_FORMAT:
	    format = "ZPixmapFormat";
	}

	String imageByteOrder;
	switch(getImageByteOrder()) {
	case LEAST_SIGNIFICANT_B_FIRST_ORDER:
	    imageByteOrder = "leastSignificantByteFirst";
	    break;
	case MOST_SIGNIFICANT_B_FIRST_ORDER:
	    imageByteOrder = "mostSignificantByteFirst";
	}

	String bitmapBitOrder;
	switch(getBitmapBitOrder()) {
	case LEAST_SIGNIFICANT_B_FIRST_ORDER:
	    bitmapBitOrder = "leastSignificantBitFirst";
	    break;
	case MOST_SIGNIFICANT_B_FIRST_ORDER:
	    bitmapBitOrder = "mostSignificantBitFirst";
	}
	

	return getClass().getName() + "[" + format +
	    ", width=" + getWidth() +
  	    ", height=" + getHeight() +
	    ", bytesPerLine=" + getBytesPerLine() +
	    ", xoffset=" + getXOffset() +
	    //	    ", visual=" + getVisual() +
  	    ", depth=" + getDepth() +
	    ", bitsPerPixel=" + getBitsPerPixel() +
	    ", bitmapUnit=" + getBitmapUnit() +
	    ", bitmapPad=" + getBitmapPad() +
	    ", byteOrder=" + imageByteOrder +
	    ", bitOrder=" + bitmapBitOrder +
	    "]";
    }

}
