package no.rwr.xutil;
import no.rwr.x.*;
import no.rwr.data.Buffers;
import java.awt.image.*;
import java.awt.color.*;

public class XImageFactory {

    Visual visual;
    ColorModel cm;
    SampleModel sm;

    int format;
    int bitsPerPixel;
    
    int dataType;

    int scanlineStride;


    /**
       Create a factory from an existing XImage.
    */
    public XImageFactory(XImage ximage, Visual visual) {
	this.visual = visual;
	this.format = ximage.getFormat();
	this.bitsPerPixel = ximage.getBitsPerPixel();


	/* FIXME: find colorModel, sampleModel and dataType for all
           ximage/visual combinations possible. */

	if (visual.getVisualClass() == Visual.VC_TRUE_COLOR) {
	    
	    // True color X11 visual == Java direct color model:

	    ColorSpace cSpace = ColorSpace.getInstance(ColorSpace.TYPE_sRGB);
	    int dataType =
		Buffers.smallestAppropriateTransferType(bitsPerPixel);

	    if (DataBuffer.getDataTypeSize(dataType) != bitsPerPixel) {
		throw new
		    UnsupportedOperationException("Unsupported number " +
						  "bits per pixel (" +
						  bitsPerPixel + "). " +
						  "No suitable data type.");
	    }

	    int   redMask = ximage.  getRedMask();
	    int greenMask = ximage.getGreenMask();
	    int  blueMask = ximage. getBlueMask();


	    // FIXME: This can be done lazily

	    cm = new DirectColorModel(cSpace,
				      bitsPerPixel,
				      redMask,
				      greenMask,
				      blueMask,
				      0, // Alpha mask
				      false, // not alpha premultiplied
				      dataType // transfer type
				      );

	    int bytesPerLine = ximage.getBytesPerLine();		  
	    this.scanlineStride = bytesPerLine * 8 / bitsPerPixel;

	    int[] masks = { redMask, greenMask, blueMask };

	    sm = new SinglePixelPackedSampleModel(dataType,
						  ximage.getWidth(),
						  ximage.getHeight(),
						  scanlineStride,
						  masks);
				      
				      
	    return;  // Done

	} 

	throw new UnsupportedOperationException("Unsupported visual (" +
						visual + "). " +
						"Unable to create " +
						"XImageFactory.");
    }
    

    public XImageFactory(Visual visual, ColorModel cm, SampleModel sm) {

	if ((sm != null) && (cm == null)) {
	    throw new IllegalArgumentException("Cannot specify a sample " +
					       "model without specifying " +
					       "a color model");
	}

	this.visual = visual;
	this.cm = cm;
	this.sm = sm;


	boolean hasSubfields = visual.hasRGBSubfields();
	
	boolean isTrueColor =
	    visual.getVisualClass() == Visual.VC_TRUE_COLOR;

	// All color models in X11 have either 1 or 3 components.
	int numColorComponents = hasSubfields ? 3 : 1;

	/* FIXME: Do further test to ensure that the visual is
	   acceptable. */

	if (cm == null) {
	    if (sm != null) {
		throw new IllegalArgumentException("Cannot specify a " +
						   "sample model " +
						   "without specifying " +
						   "a color model");
	    }
	    return; // All OK
	}

	if (numColorComponents != cm.getNumColorComponents()) {
	    // FIXME: create a special ImageFormatMismatchException.
	    throw 
		new IllegalArgumentException("Visual " + visual +
					     " expects a different " +
					     "number of components than " +
					     "provided by color model " +
					     cm + " (" + numColorComponents +
					     "!=" + 
					     cm.getNumColorComponents() +
					     ")");
	}

	if (cm instanceof ComponentColorModel) {
	    ComponentColorModel ccm = (ComponentColorModel) cm;
	    ComponentSampleModel csm = (ComponentSampleModel) sm;
	
	    int depth = visual.getDepth();
	    int[] componentSizes = ccm.getComponentSize();
	    int pixelSize = 0;
	    for (int b=0; b<numColorComponents; b++) 
		pixelSize += componentSizes[b];
	    
	    if (depth != pixelSize) {
		throw 
		    new IllegalArgumentException("Visual " + visual +
						 " expects a different " +
						 "depth than provided " +
						 "by color model " +
						 cm + " (" + depth +
						 "!=" +  pixelSize + ")");
	    }

	    dataType = sm.getDataType();

	    bitsPerPixel =
		DataBuffer.getDataTypeSize(dataType);
	    scanlineStride = csm.getScanlineStride();
	    
	    // FIXME: don't assume
	    format = XImage.ZPIXMAP_FORMAT;
	
	    
	    //boolean leastSignificantByteFirst = true;
	    //boolean mostSignificantByteFirst = true;
	    
	    //int[] bandOffsets = csm.getBandOffsets();
	
	} else {
	    throw new UnsupportedOperationException("do not know how to " +
						    "handle color model " +
						    cm.getClass());
	}
    }

    public XImage createXImage(int width, int height) {
	// FIXME: implement

	/* FIXME: We have a problem here. The data buffer is needed in
	   the Image, but we only return an XImage. How do we avoid
	   loosing the reference to this data buffer?

	   Maybe we shouldn't create a data buffer here at all, but
	   leave this job up to the Image class? */

	return null;
    }

    public XImage createXImage(Raster raster) {

	int xoffset = 0;
	int width = raster.getWidth();
	int height = raster.getHeight();
	int bitmapPad = 32; // FIXME: don't hardcode this

	DataBuffer dataB = raster.getDataBuffer();

	if (dataType != dataB.getDataType()) {
	    throw new IllegalArgumentException("Data type mismatch between " +
					       "factory (" + dataType + ") " +
					       "and raster (" + 
					       dataB.getDataType() + ").");
	}

	int dataType = dataB.getDataType();
	int bitsPerUnit = DataBuffer.getDataTypeSize(dataType);
	int bytesPerLine = scanlineStride * bitsPerUnit / 8;
	
	XImage ximage = new XImage(visual, visual.getDepth(), format,
				   xoffset, width, height, bitmapPad,
				   bytesPerLine);
	
	int x = raster.getMinX() - raster.getSampleModelTranslateX();
	int y = raster.getMinY() - raster.getSampleModelTranslateY();

	int offset = 
	    ((bitsPerPixel*x)/bitsPerUnit) +   // from sample model
	    scanlineStride*y;                  // from sample model
	attachData(ximage, dataB, offset);
	return ximage;
    }

    public DataBuffer createDataBuffer(XImage ximage) {
	return getSampleModel().createDataBuffer();
    }
    
    /**
       @param offset Offset to data. The given offset does not include
       data buffer offset, which will also be added.  */
    public static void
	attachData(XImage ximage, DataBuffer dataB, int offset) {
	offset += dataB.getOffset();
	switch (dataB.getDataType()) {
	case DataBuffer.TYPE_BYTE:
	    ximage.setData(((DataBufferByte) dataB).getData(), offset);
	    break;
	case DataBuffer.TYPE_USHORT:
	    ximage.setData(((DataBufferUShort) dataB).getData(), offset);
	    break;
	default:
	    throw
		new UnsupportedOperationException("Do not know how to " +
						  "set data for data " +
						  "type " +
						  dataB.getDataType());
	}
    }
    




    public Visual getVisual() {
	return visual;
    }

    public ColorModel getColorModel() {
	// FIXME: If null, generate appropriate color model.
	return cm;
    }

    public SampleModel getSampleModel() {
	// FIXME: If null, generate appropriate sample model.
	return sm;
    }


    public String toString() {
	return getClass().getName() + "[" +
	    "visual=" + visual +
  	    ", colorModel=" + cm +
  	    ", sampleModel=" + sm +
  	    ", format=" + format +
  	    ", bitsPerPixel=" + bitsPerPixel +
  	    ", dataType=" + dataType +
  	    ", scanlineStride=" + scanlineStride + "]";
    }

}
